import datetime as dt
import os

import numpy as np
import pandas as pd
import xarray as xr
import yaml

from .shared import (
    resample,
    get_stable_fluxes,
    get_vertical_transport,
    get_grid_info,
    join_levels,
    repeat_upper_level,
    get_new_target_levels,
    interpolate,
)


# Set constants
g = 9.80665  # [m/s2]
density_water = 1000  # [kg/m3]

# Read case configuration
with open("cases/ec-earth.yaml") as f:
    config = yaml.safe_load(f)

# Parse input from config file
# Reassignment not strictly needed but improves readability for often used vars
input_folder = config["input_folder"]
name_of_run = config["name_of_run"]
divt = config["divt"]
count_time = config["count_time"]
latnrs = np.arange(config["latnrs"])
lonnrs = np.arange(config["lonnrs"])


def _get_input_data(variable, date, latnrs, lonnrs):
    """Get input data for variable."""
    filename = f"{name_of_run}{variable}_{date.year}{date.month:02d}_NH.nc"
    filepath = os.path.join(config["input_folder"], filename)
    return (
        xr.open_dataset(filepath)
        .sel(time=date.strftime("%Y%m%d"))
        .isel(lat=latnrs, lon=lonnrs)
    )


def get_input_data(variable, date, latnrs, lonnrs):
    """Workaround for keeping original timestamps crossing midnight."""
    day1 = _get_input_data(variable, date, latnrs, lonnrs)
    day2 = _get_input_data(variable, date + dt.timedelta(days=1), latnrs, lonnrs)
    days = xr.concat([day1, day2], dim="time")
    if len(days.time) < 9:
        return days.isel(time=slice(0, 5)).isel(lat=latnrs, lon=lonnrs)
    else:
        return days.isel(time=slice(0, 10)).isel(lat=latnrs, lon=lonnrs)


datelist = pd.date_range(
    start=config["start_date"], end=config["end_date"], freq="d", inclusive="left"
)

for date in datelist:
    print(date)

    # Load data
    time_3h = slice(1, None, 2)
    times_lnsp = slice(None, None, 2)  # TODO should be slice(1, None, 2) ??

    q = get_input_data("Q", date, latnrs, lonnrs).Q
    u = get_input_data("U", date, latnrs, lonnrs).U
    v = get_input_data("V", date, latnrs, lonnrs).V
    q2m = get_input_data("Q2M", date, latnrs, lonnrs).Q2M.isel(time=time_3h)
    lnsp = get_input_data("LNSP", date, latnrs, lonnrs).LNSP.isel(time=times_lnsp)
    u10 = get_input_data("U10", date, latnrs, lonnrs).U10M.isel(time=time_3h)
    v10 = get_input_data("V10", date, latnrs, lonnrs).V10M.isel(time=time_3h)

    sp = np.exp(lnsp)  # log(sp) --> sp (Pa)

    # Create pressure array with the same dimensions as u, q, and v
    p = u.lev.broadcast_like(u)

    # Get the most out of the data we have
    u = join_levels(u, u10)  # bottom level will be set to 1000 00 Pa
    v = join_levels(v, v10)
    q = join_levels(q, q2m)
    p = join_levels(p, sp.squeeze())

    u = repeat_upper_level(u)  # top level will be set to 100 00 Pa
    v = repeat_upper_level(v)
    q = repeat_upper_level(q, fill_value=0)
    p = repeat_upper_level(p, fill_value=0)

    # Mask data where the pressure level is higher than sp - 1000
    # as there is no valid data at those points
    # and we don't need a second layer that is very close to the surface
    mask = p > sp.values - 1000
    mask[:, 0, :, :] = False  # don't mask bottom (surface pressure values)
    mask[:, -1, :, :] = False  # don't mask top ("ghost cells"?)

    # TODO convert to nan instead of numpy masked array?
    # u_masked = u.where(mask)
    u = np.ma.masked_array(u, mask=mask)
    v = np.ma.masked_array(v, mask=mask)
    q = np.ma.masked_array(q, mask=mask)
    p = np.ma.masked_array(p, mask=mask)

    ####
    # Get grid info
    dummy = xr.open_dataset(config["land_sea_mask"])
    lat = dummy.LAT.values[444:711][::-1]  # [degrees north]
    lon = dummy.XAS.values[934:1378][::-1]  # [degrees east]
    a_gridcell, l_ew_gridcell, l_mid_gridcell = get_grid_info(lat, lon)

    # Reverse engineered ~ model level 47 which corresponds with about ~800 hPa
    p_boundary = 0.72878581 * sp.values + 7438.803223
    new_pressure_levels = get_new_target_levels(sp, p_boundary, n_levels=40)

    print("before interpolation loop", dt.datetime.now().time())
    uq_boundaries = interpolate(u * q, p, new_pressure_levels, type="cubic")
    vq_boundaries = interpolate(v * q, p, new_pressure_levels, type="cubic")
    q_boundaries = interpolate(q, p, new_pressure_levels, type="linear")

    # Integrate specific humidity to get the (total) column water (vapor) and calculate horizontal moisture fluxes
    q_midpoints = 0.5 * (q_boundaries[:, 1:, :, :] + q_boundaries[:, :-1, :, :])
    uq_midpoints = 0.5 * (uq_boundaries[:, 1:, :, :] + uq_boundaries[:, :-1, :, :])
    vq_midpoints = 0.5 * (vq_boundaries[:, 1:, :, :] + vq_boundaries[:, :-1, :, :])
    # for p we do not calculate the mean pressure but the pressure difference between two levels!
    p_diff = np.maximum(
        0, new_pressure_levels[:, :-1, :, :] - new_pressure_levels[:, 1:, :, :]
    )  # the maximum statement is necessary to avoid negative humidity values

    # eastward and northward fluxes
    fa_e_p = uq_midpoints * p_diff / g
    fa_n_p = vq_midpoints * p_diff / g

    column_water_vapor = q_midpoints * p_diff / g  # [kg/m2]

    # summed over the vertical
    total_column_water_vapor = np.squeeze(np.sum(column_water_vapor, 1))

    mask = np.where(new_pressure_levels > p_boundary, 1.0, 0.0)

    vapor_lower = np.sum(mask[:, :-1, :, :] * q_midpoints * p_diff / g, axis=1)
    vapor_upper = np.sum((1 - mask[:, :-1, :, :]) * q_midpoints * p_diff / g, axis=1)

    fa_e_lower = np.sum(mask[:, :-1, :, :] * fa_e_p, axis=1)  # kg*m-1*s-1
    fa_n_lower = np.sum(mask[:, :-1, :, :] * fa_n_p, axis=1)  # kg*m-1*s-1
    fa_e_upper = np.sum((1 - mask[:, :-1, :, :]) * fa_e_p, axis=1)  # kg*m-1*s-1
    fa_n_upper = np.sum((1 - mask[:, :-1, :, :]) * fa_n_p, axis=1)  # kg*m-1*s-1

    vapor_total = vapor_upper + vapor_lower

    check = np.sum(total_column_water_vapor - vapor_total)
    print(f"Check calculation water vapor, this value should be zero: {check}")

    # water volumes
    w_upper = vapor_upper * a_gridcell[None, ...] / density_water  # m3
    w_lower = vapor_lower * a_gridcell[None, ...] / density_water  # m3

    ## Get E and P
    # (accumulated after the forecast at 00.00 and 12.00 by steps of 3 hours in time
    evaporation = get_input_data("EVAP", date, latnrs, lonnrs).E  # m
    precipitation = get_input_data("TP", date, latnrs, lonnrs).TP  # m

    # change sign convention to all positive, transfer negative (originally positive) values of evap to precip
    precipitation = np.maximum(precipitation + np.maximum(evaporation, 0), 0)
    evaporation = np.abs(np.minimum(evaporation, 0))

    # calculate volumes
    evap = (evaporation * a_gridcell).values
    precip = (precipitation * a_gridcell).values

    # put data on a smaller time step
    evap = resample(evap, divt / 2, count_time, method="bfill")[:-1]
    precip = resample(precip, divt / 2, count_time, method="bfill")[:-1]
    w_upper = resample(w_upper, divt, count_time, method="interp")
    w_lower = resample(w_lower, divt, count_time, method="interp")
    fa_e_upper = resample(fa_e_upper, divt, count_time, method="interp")[:-1]
    fa_e_lower = resample(fa_e_lower, divt, count_time, method="interp")[:-1]
    fa_n_upper = resample(fa_n_upper, divt, count_time, method="interp")[:-1]
    fa_n_lower = resample(fa_n_lower, divt, count_time, method="interp")[:-1]

    # convert to m3   [kg*m^-1 * s^-1 * s * m * kg^-1 * m^3] = [m3]
    total_seconds = config["timestep"] / config["divt"]
    fa_e_upper *= total_seconds * (l_ew_gridcell / density_water)
    fa_e_lower *= total_seconds * (l_ew_gridcell / density_water)
    fa_n_upper *= total_seconds * (l_mid_gridcell[None, :, None] / density_water)
    fa_n_lower *= total_seconds * (l_mid_gridcell[None, :, None] / density_water)

    # stabilize horizontal fluxes
    fa_e_upper, fa_n_upper = get_stable_fluxes(fa_e_upper, fa_n_upper, w_upper)
    fa_e_lower, fa_n_lower = get_stable_fluxes(fa_e_lower, fa_n_lower, w_lower)

    # determine the vertical moisture flux
    fa_vert = get_vertical_transport(
        fa_e_upper, fa_e_lower, fa_n_upper, fa_n_lower, evap, precip, w_upper, w_lower
    )

    # Save preprocessed data
    filename = f"{date.strftime('%Y-%m-%d')}_fluxes_storages.nc"
    output_path = os.path.join(config["interdata_folder"], filename)
    xr.Dataset(
        {  # TODO: would be nice to add coordinates and units as well
            "fa_e_upper": (["time", "lat", "lon"], fa_e_upper),
            "fa_n_upper": (["time", "lat", "lon"], fa_n_upper),
            "fa_e_lower": (["time", "lat", "lon"], fa_e_lower),
            "fa_n_lower": (["time", "lat", "lon"], fa_n_lower),
            "evap": (["time", "lat", "lon"], evap),
            "precip": (["time", "lat", "lon"], precip),
            "w_upper": (["time2", "lat", "lon"], w_upper),  # note different time
            "w_lower": (["time2", "lat", "lon"], w_lower),  # note different time
            "fa_vert": (["time", "lat", "lon"], fa_vert),
        }
    ).to_netcdf(output_path)



from pathlib import Path

import click
import numpy as np
import pandas as pd
import xarray as xr
import yaml
from wam2layers.analysis.checks import check_input
from wam2layers.preprocessing.shared import (accumulation_to_flux,
                                                    calculate_humidity,
                                                    insert_level, interpolate,
                                                    sortby_ndarray)


def load_data(variable, date, config):
    """Load data for given variable and date."""
    prefix = config["filename_prefix"]
    # Include midnight of the next day (if available)
    extra = date + pd.Timedelta(days=1)

    if variable in ["u", "v", "q", "t"] and config["level_type"] == "model_levels":
        prefix += "_ml"

    filepath = Path(config["input_folder"]) / f"{prefix}_{variable}.nc"
    da = xr.open_dataset(filepath)[variable].sel(time=slice(date, extra))

    if "lev" in da.coords:
        da = da.rename(lev="level")

    if variable in ["u", "v", "q", "t"] and isinstance(config["levels"], list):
        return da.sel(level=config["levels"])

    return da.sel(time=slice(date, extra))


def preprocess_precip_and_evap(date, config):
    """Load and pre-process precipitation and evaporation."""
    # All incoming units are accumulations (in m) since previous time step
    evap = load_data("e", date, config)
    cp = load_data("cp", date, config)  # convective precipitation
    lsp = load_data("lsp", date, config)  # large scale precipitation
    precip = (cp + lsp)

    # Transfer negative (originally positive) values of evap to precip
    precip = np.maximum(precip, 0) + np.maximum(evap, 0)
    precip = precip

    # Change sign convention to all positive,
    evap = np.abs(np.minimum(evap, 0))

    precip = accumulation_to_flux(precip)
    evap = accumulation_to_flux(evap)

    return precip, evap


def get_edges(era5_modellevels):
    """Get the values of a and b at the edges of a subset of ERA5 modellevels."""
    # Load a and b coefficients
    table = Path(__file__).parent / "tableERA5model_to_pressure.csv"
    df = pd.read_csv(table)
    a = df["a [Pa]"].to_xarray().rename(index="level")
    b = df["b"].to_xarray().rename(index="level")

    # Calculate a and b at mid levels (model levels)
    a_full = ((a[1:] + a[:-1].values) / 2.0).sel(level=era5_modellevels)
    b_full = ((b[1:] + b[:-1].values) / 2.0).sel(level=era5_modellevels)

    # Interpolate to get parameters at edges between selected levels
    a_edge = xr.concat([a[0], (a_full[1:].values + a_full[:-1]) / 2, a[-1]], dim="level")
    b_edge = xr.concat([b[0], (b_full[1:].values + b_full[:-1]) / 2, b[-1]], dim="level")

    return a_edge, b_edge


def get_dp_modellevels(sp, modellevels):
    """Calculate pressure jump over subset of era5 modellevels."""
    if modellevels == "all":
        modellevels = list(range(138))

    sp_broadcast = sp.expand_dims({"level": modellevels}, axis=1)
    a_edge, b_edge = get_edges(modellevels)
    p_edges = a_edge + b_edge * sp_broadcast  # in Pa

    # calculate the difference between the pressure levels
    dp_modellevels = p_edges.diff(dim="level")  # in Pa

    return dp_modellevels


def get_dp_pressurelevels(q, u, v, ps, qs, us, vs):
    """Get dp with consistent u, v, q for ERA5 pressure level data."""
    p = u.level.broadcast_like(u) * 100  # Pa

    # Insert top of atmosphere values
    # Assume wind at top same as values at lowest pressure, humidity at top 0
    u = insert_level(u, u.isel(level=0), 0)
    v = insert_level(v, v.isel(level=0), 0)
    q = insert_level(q, 0, 0)
    p = insert_level(p, 0, 0)

    # Insert surface level values (at a high dummy pressure value)
    u = insert_level(u, us, 110000)
    v = insert_level(v, vs, 110000)
    q = insert_level(q, qs, 110000)
    p = insert_level(p, ps, 110000)

    # Sort arrays by pressure (ascending)
    u.values = sortby_ndarray(u.values, p.values, axis=1)
    v.values = sortby_ndarray(v.values, p.values, axis=1)
    q.values = sortby_ndarray(q.values, p.values, axis=1)
    p.values = sortby_ndarray(p.values, p.values, axis=1)

    # Insert boundary level values (at a ridiculous dummy pressure value)
    p_boundary = 0.72878581 * np.array(ps) + 7438.803223
    u = insert_level(u, interpolate(p_boundary, p, u), 150000)
    v = insert_level(v, interpolate(p_boundary, p, v), 150000)
    q = insert_level(q, interpolate(p_boundary, p, q), 150000)
    p = insert_level(p, p_boundary, 150000)

    # Sort arrays by pressure once more (ascending)
    u.values = sortby_ndarray(u.values, p.values, axis=1)
    v.values = sortby_ndarray(v.values, p.values, axis=1)
    q.values = sortby_ndarray(q.values, p.values, axis=1)
    p.values = sortby_ndarray(p.values, p.values, axis=1)

    # Reset level coordinate as its values have become meaningless
    nlev = u.level.size
    u = u.assign_coords(level=np.arange(nlev))
    v = v.assign_coords(level=np.arange(nlev))
    q = q.assign_coords(level=np.arange(nlev))
    p = p.assign_coords(level=np.arange(nlev))

    # Calculate pressure jump
    dp = p.diff("level")
    assert np.all(dp > 0), "Pressure levels should increase monotonically"

    # Interpolate to midpoints
    midpoints = 0.5 * (u.level.values[1:] + u.level.values[:-1])
    dp = dp.assign_coords(level=midpoints)
    u = u.interp(level=midpoints)
    v = v.interp(level=midpoints)
    q = q.interp(level=midpoints)
    p = p.interp(level=midpoints)

    # mask values below surface
    above_surface = p < np.array(ps)[:, None, :, :]
    u = u.where(above_surface)
    v = v.where(above_surface)
    q = q.where(above_surface)
    p = p.where(above_surface)

    return dp, p, q, u, v, p_boundary


def parse_config(config_file):
    """Read and parse case configuration file."""
    with open(config_file) as f:
        config = yaml.safe_load(f)

    # Create the preprocessed data folder if it does not exist yet
    config["output_dir"] = Path(config["preprocessed_data_folder"]).expanduser()
    config["output_dir"].mkdir(exist_ok=True, parents=True)

    config["datelist"] = pd.date_range(
        start=config["preprocess_start_date"],
        end=config["preprocess_end_date"],
        freq="d",
        inclusive="left",
    )
    return config


def prep_experiment(config_file):
    """Pre-process all data for a given config file.

    This function expects the following configuration settings:

    - preprocess_start_date: formatted as YYYYMMDD, e.g. '20210701'
    - preprocess_end_date: formatted as YYYYMMDD, e.g. '20210716'
    - level_type: either "pressure_levels" or "model_levels"
    - levels: "all" or a list of integers with the desired (model or pressure)
      levels.
    - input_folder: path where raw era5 input data can be found, e.g.
      /home/peter/WAM2layers/era5_2021
    - preprocessed_data_folder: path where preprocessed data should be stored.
      This directory will be created if it does not exist. E.g.
      /home/peter/WAM2layers/preprocessed_data_2021
    - filename_prefix: Fixed part of filename. This function will infer the
      variable name and add _ml for model level data. E.g. with prefix =
      "FloodCase_202107" this function will be able to find
      FloodCase_202107_ml_u.nc or FloodCase_202107_u.nc and
      FloodCase_202107_sp.nc
    """
    config = parse_config(config_file)
    for date in config["datelist"]:
        print(date)
        precip, evap = preprocess_precip_and_evap(date, config)

        # 4d fields
        levels = config["levels"]
        q = load_data("q", date, config)  # in kg kg-1
        u = load_data("u", date, config)  # in m/s
        v = load_data("v", date, config)  # in m/s
        sp = load_data("sp", date, config)  # in Pa

        if config["level_type"] == "model_levels":
            dp = get_dp_modellevels(sp, levels)

        if config["level_type"] == "pressure_levels":
            d2m = load_data("d2m", date, config)  # Dew point in K
            q2m = calculate_humidity(d2m, sp)  # kg kg-1
            u10 = load_data("u10", date, config)  # in m/s
            v10 = load_data("v10", date, config)  # in m/s
            dp, p, q, u, v, pb = get_dp_pressurelevels(q, u, v, sp, q2m, u10, v10)

        # Calculate column water vapour
        g = 9.80665  # gravitational accelleration [m/s2]
        cwv = q * dp / g  # (kg/m2)

        try:
            # Calculate column water instead of column water vapour
            tcw = load_data("tcw", date, config)  # kg/m2
            cw = (tcw / cwv.sum(dim="level")) * cwv  # column water (kg/m2)
            # TODO: warning if cw >> cwv
        except FileNotFoundError:
            # Fluxes will be calculated based on the column water vapour
            cw = cwv

        # Integrate fluxes and states to upper and lower layer
        if config["level_type"] == "model_levels":
            # TODO: Check if this is a reasonable choice for boundary
            boundary = 111
            lower_layer = dp.level > boundary
            upper_layer = ~lower_layer
        if config["level_type"] == "pressure_levels":
            upper_layer = p < pb[:, None, :, :]
            lower_layer = pb[:, None, :, :] < p

        # Vertically integrate state over two layers
        s_lower = cw.where(lower_layer).sum(dim="level")
        s_upper = cw.where(upper_layer).sum(dim="level")

        # Determine the fluxes
        fx = u * cw  # eastward atmospheric moisture flux (kg m-1 s-1)
        fy = v * cw  # northward atmospheric moisture flux (kg m-1 s-1)

        # Vertically integrate fluxes over two layers
        fx_lower = fx.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fy_lower = fy.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fx_upper = fx.where(upper_layer).sum(dim="level")  # kg m-1 s-1
        fy_upper = fy.where(upper_layer).sum(dim="level")  # kg m-1 s-1

        # Combine everything into one dataset
        ds = xr.Dataset(
            {
                "fx_upper": fx_upper.assign_attrs(units="kg m-1 s-1"),
                "fy_upper": fy_upper.assign_attrs(units="kg m-1 s-1"),
                "fx_lower": fx_lower.assign_attrs(units="kg m-1 s-1"),
                "fy_lower": fy_lower.assign_attrs(units="kg m-1 s-1"),
                "s_upper": s_upper.assign_attrs(units="kg m-2"),
                "s_lower": s_lower.assign_attrs(units="kg m-2"),
                "evap": evap,
                "precip": precip,
            }
        )

        # Verify that the data meets all the requirements for the model
        check_input(ds)

        # Save preprocessed data
        filename = f"{date.strftime('%Y-%m-%d')}_fluxes_storages.nc"
        output_path = config["output_dir"] / filename
        ds.to_netcdf(output_path)

################################################################################
# To run this script interactively in e.g. Spyder, uncomment the following line:
# prep_experiment("../../cases/era5_2021.yaml")
################################################################################

###########################################################################
# The code below makes it possible to run wam2layers from the command line:
# >>> python backtrack.py path/to/cases/era5_2021.yaml
# or even:
# >>> wam2layers backtrack path/to/cases/era5_2021.yaml
###########################################################################


@click.command()
@click.argument('config_file', type=click.Path(exists=True))
def cli(config_file):
    """Preprocess ERA5 data for WAM2layers tracking experiments.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - python path/to/preprocessing/era5.py path/to/cases/era5_2021.yaml
        - wam2layers preprocess era5 path/to/cases/era5_2021.yaml
    """
    prep_experiment(config_file)


if __name__ == "__main__":
    cli()



"""Command-line interface for the WAM2layers model."""
import click
from wam2layers.tracking.backtrack import cli as backtrack_cli
from wam2layers.preprocessing.cli import cli as preproc_cli

# Visualization requires optional dependencies that may not be installed.
try:
    from wam2layers.analysis.visualization import cli as visualize_cli
    include_visualize = True
except ImportError:
    include_visualize = False
    click.echo(
        "To enable visualization options, install matplotlib and cartopy. "
    )


@click.group()
def cli():
    """Command line interface to WAM2layers."""
    pass


cli.add_command(preproc_cli, name="preprocess")
cli.add_command(backtrack_cli, name="backtrack")

# Visualization requires optional dependencies that may not be installed.
if include_visualize:
    cli.add_command(visualize_cli, name="visualize")



from pathlib import Path

import click
import numpy as np
import pandas as pd
import xarray as xr
import yaml
from wam2layers.preprocessing.shared import get_grid_info


def parse_config(config_file):
    """Load and parse config file into dictionary."""
    with open(config_file) as f:
        config = yaml.safe_load(f)

    # Add datelist
    config["datelist"] = pd.date_range(
        start=config["track_start_date"],
        end=config["track_end_date"],
        freq="d",
        inclusive="left",
    )

    # Parse directories with pathlib
    config["preprocessed_data_folder"] = Path(config["preprocessed_data_folder"]).expanduser()
    config["output_folder"] = Path(config["output_folder"]).expanduser() / "backtrack"

    # Check if input dir exists
    if not config["preprocessed_data_folder"].exists():
        raise ValueError(
            "Please create the preprocessed_data_folder before running the script"
        )

    # Create output dir if it doesn't exist yet
    if not config["output_folder"].exists():
        config["output_folder"].mkdir(parents=True)

    return config


def time_in_range(start, end, current):
    """Returns whether current is in the range [start, end]"""
    return start <= current <= end


def load_region(config):
    # TODO: make variable name more generic
    return xr.open_dataset(config["region"]).source_region


def input_path(date, config):
    input_dir = config["preprocessed_data_folder"]
    return f"{input_dir}/{date.strftime('%Y-%m-%d')}_fluxes_storages.nc"


def output_path(date, config):
    output_dir = config["output_folder"]
    return f"{output_dir}/{date.strftime('%Y-%m-%d')}_s_track.nc"


def to_edges_zonal(fx, periodic_boundary=False):
    """Define the horizontal fluxes over the east/west boundaries."""
    fe = np.zeros_like(fx)
    fe[:, :-1] = 0.5 * (fx[:, :-1] + fx[:, 1:])
    if periodic_boundary:
        fe[:, -1] = 0.5 * (fx[:, -1] + fx[:, 0])

    # find out where the positive and negative fluxes are
    f_pos = np.ones_like(fx)
    f_pos[fe < 0] = 0
    f_neg = f_pos - 1

    # separate directions west-east (all positive numbers)
    fe_we = fe * f_pos
    fe_ew = fe * f_neg

    # fluxes over the western boundary
    fw_we = look_west(fe_we)
    fw_ew = look_west(fe_ew)

    return fe_we, fe_ew, fw_we, fw_ew


def to_edges_meridional(fy):
    """Define the horizontal fluxes over the north/south boundaries."""
    fn = np.zeros_like(fy)
    fn[1:, :] = 0.5 * (fy[:-1, :] + fy[1:, :])

    # find out where the positive and negative fluxes are
    fn_pos = np.ones_like(fn)
    fn_pos[fn < 0] = 0  # invalid value encountered in less
    fn_neg = fn_pos - 1

    # separate directions south-north (all positive numbers)
    fn_sn = fn * fn_pos
    fn_ns = fn * fn_neg

    # fluxes over the southern boundary
    fs_sn = look_south(fn_sn)
    fs_ns = look_south(fn_ns)

    return fn_sn, fn_ns, fs_sn, fs_ns


def look_north(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, 1, axis=-2)


def look_south(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, -1, axis=-2)


def look_east(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, -1, axis=-1)


def look_west(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, 1, axis=-1)


def split_vertical_flux(Kvf, fv):
    f_downward = np.zeros_like(fv)
    f_upward = np.zeros_like(fv)
    f_downward[fv >= 0] = fv[fv >= 0]
    f_upward[fv <= 0] = fv[fv <= 0]
    f_upward = np.abs(f_upward)

    # include the vertical dispersion
    if Kvf != 0:
        f_upward = (1.0 + Kvf) * f_upward
        f_upward[fv >= 0] = fv[fv >= 0] * Kvf
        f_downward = (1.0 + Kvf) * f_downward
        f_downward[fv <= 0] = np.abs(fv[fv <= 0]) * Kvf

    return f_downward, f_upward


def resample(ds, target_freq):
    """Increase time resolution; states at midpoints, fluxes at the edges."""
    time = ds.time.values
    newtime_states = pd.date_range(time[0], time[-1], freq=target_freq)
    newtime_fluxes = newtime_states[:-1] + pd.Timedelta(target_freq) / 2

    states = ds[['s_upper', 's_lower']].interp(time=newtime_states)
    fluxes = ds[['fx_upper', 'fx_lower', 'fy_upper', 'fy_lower', 'precip', 'evap']].interp(time=newtime_fluxes)
    return fluxes, states


def change_units(fluxes, states, target_freq):
    """Change units to m3.
    Multiply by edge length or area to get flux in m3
    Multiply by time to get accumulation instead of flux
    Divide by density of water to go from kg to m3
    """
    density = 1000  # [kg/m3]
    a, ly, lx = get_grid_info(fluxes)

    total_seconds = pd.Timedelta(target_freq).total_seconds()
    fluxes["fx_upper"] *= total_seconds / density * ly
    fluxes["fx_lower"] *= total_seconds / density * ly
    fluxes["fy_upper"] *= total_seconds / density * lx[None, :, None]
    fluxes["fy_lower"] *= total_seconds / density * lx[None, :, None]
    fluxes["evap"] *= total_seconds / density * a[None, :, None]
    fluxes["precip"] *= total_seconds / density * a[None, :, None]
    states["s_upper"] *= a[None, :, None] / density
    states["s_lower"] *= a[None, :, None] / density

    for variable in fluxes.data_vars:
        fluxes[variable] = fluxes[variable].assign_attrs(units="m**3")

    for variable in states.data_vars:
        states[variable] = states[variable].assign_attrs(units="m**3")


def stabilize_fluxes(fluxes, states):
    """Stabilize the outfluxes / influxes.

    During the reduced timestep the water cannot move further than 1/x * the
    gridcell, In other words at least x * the reduced timestep is needed to
    cross a gridcell.
    """
    for level in ["upper", "lower"]:
        fx = fluxes["fx_" + level]
        fy = fluxes["fy_" + level]
        s = states["s_" + level]

        fx_abs = np.abs(fx)
        fy_abs = np.abs(fy)
        ft_abs = fx_abs + fy_abs

        fx_corrected = 1/2 * fx_abs / ft_abs * s[:-1, :, :].values
        fx_stable = np.minimum(fx_abs, fx_corrected)

        fy_corrected = 1/2 * fy_abs / ft_abs * s[:-1, :, :].values
        fy_stable = np.minimum(fy_abs, fy_corrected)

        # Get rid of any nan values
        fx_stable.fillna(0)
        fy_stable.fillna(0)

        # Re-instate the sign
        fluxes["fx_"+ level] = np.sign(fx) * fx_stable
        fluxes["fy_"+ level] = np.sign(fy) * fy_stable


def convergence(fx, fy):
    # Note: latitude decreasing, hence positive fy gradient is convergence
    return np.gradient(fy, axis=-2) - np.gradient(fx, axis=-1)


def calculate_fv(fluxes, states, kvf, periodic):
    """Calculate the vertical fluxes.

    Note: fluxes are given at temporal midpoints between states.
    """
    s_total = states.s_upper + states.s_lower
    s_rel_upper = (states.s_upper / s_total).interp(time=fluxes.time)
    s_rel_lower = (states.s_lower / s_total).interp(time=fluxes.time)

    tendency_upper = convergence(fluxes.fx_upper, fluxes.fy_upper) - fluxes.precip.values * s_rel_upper
    tendency_lower = convergence(fluxes.fx_upper, fluxes.fy_upper) - fluxes.precip.values * s_rel_lower + fluxes.evap

    residual_upper = states.s_upper.diff("time").values - tendency_upper
    residual_lower = states.s_lower.diff("time").values - tendency_lower

    # compute the resulting vertical moisture flux; the vertical velocity so
    # that the new residual_lower/s_lower = residual_upper/s_upper (positive downward)
    fv = s_rel_lower * (residual_upper + residual_lower) - residual_lower

    # stabilize the outfluxes / influxes; during the reduced timestep the
    # vertical flux can maximally empty/fill 1/x of the top or down storage
    stab = 1.0 / (kvf + 1.0)
    flux_limit = np.minimum(states.s_upper, states.s_lower).interp(time=fluxes.time)
    fv_stable = np.minimum(np.abs(fv), stab * flux_limit)

    # Reinstate the sign
    return np.sign(fv) * fv_stable


def backtrack(
    date,
    fluxes,
    states,
    s_track_upper,
    s_track_lower,
    region,
    config,
):

    # Unpack preprocessed data
    fx_upper = fluxes["fx_upper"].values
    fy_upper = fluxes["fy_upper"].values
    fx_lower = fluxes["fx_lower"].values
    fy_lower = fluxes["fy_lower"].values
    evap = fluxes["evap"].values
    precip = fluxes["precip"].values
    f_vert = fluxes["f_vert"].values
    s_upper = states["s_upper"].values
    s_lower = states["s_lower"].values

    # Allocate arrays for daily accumulations
    ntime, nlat, nlon = fx_upper.shape

    s_track_upper_mean = np.zeros((nlat, nlon))
    s_track_lower_mean = np.zeros((nlat, nlon))
    e_track = np.zeros((nlat, nlon))

    north_loss = np.zeros(nlon)
    south_loss = np.zeros(nlon)
    east_loss = np.zeros(nlat)
    west_loss = np.zeros(nlat)

    # Only track the precipitation at certain dates
    if (
        time_in_range(
            config["event_start_date"],
            config["event_end_date"],
            date.strftime("%Y%m%d"),
        )
        == False
    ):
        precip = precip * 0

    # Sa calculation backward in time
    for t in reversed(range(ntime)):
        P_region = region * precip[t]
        s_total = s_upper[t+1] + s_lower[t+1]

        # separate the direction of the vertical flux and make it absolute
        f_downward, f_upward = split_vertical_flux(config["kvf"], f_vert[t])

        # Determine horizontal fluxes over the grid-cell boundaries
        f_e_lower_we, f_e_lower_ew, f_w_lower_we, f_w_lower_ew = to_edges_zonal(
            fx_lower[t]
        )
        f_e_upper_we, f_e_upper_ew, f_w_upper_we, f_w_upper_ew = to_edges_zonal(
            fx_upper[t]
        )

        (
            fy_n_lower_sn,
            fy_n_lower_ns,
            fy_s_lower_sn,
            fy_s_lower_ns,
        ) = to_edges_meridional(fy_lower[t])
        (
            fy_n_upper_sn,
            fy_n_upper_ns,
            fy_s_upper_sn,
            fy_s_upper_ns,
        ) = to_edges_meridional(fy_upper[t])

        # Short name for often used expressions
        s_track_relative_lower = (
            s_track_lower / s_lower[t+1]
        )  # fraction of tracked relative to total moisture
        s_track_relative_upper = s_track_upper / s_upper[t+1]
        inner = np.s_[1:-1, 1:-1]

        # Actual tracking (note: backtracking, all terms have been negated)
        s_track_lower[inner] += (
            + f_e_lower_we * look_east(s_track_relative_lower)
            + f_w_lower_ew * look_west(s_track_relative_lower)
            + fy_n_lower_sn * look_north(s_track_relative_lower)
            + fy_s_lower_ns * look_south(s_track_relative_lower)
            + f_upward * s_track_relative_upper
            - f_downward * s_track_relative_lower
            - fy_s_lower_sn * s_track_relative_lower
            - fy_n_lower_ns * s_track_relative_lower
            - f_e_lower_ew * s_track_relative_lower
            - f_w_lower_we * s_track_relative_lower
            + P_region * (s_lower[t+1] / s_total)
            - evap[t] * s_track_relative_lower
        )[inner]

        s_track_upper[inner] += (
            + f_e_upper_we * look_east(s_track_relative_upper)
            + f_w_upper_ew * look_west(s_track_relative_upper)
            + fy_n_upper_sn * look_north(s_track_relative_upper)
            + fy_s_upper_ns * look_south(s_track_relative_upper)
            + f_downward * s_track_relative_lower
            - f_upward * s_track_relative_upper
            - fy_s_upper_sn * s_track_relative_upper
            - fy_n_upper_ns * s_track_relative_upper
            - f_w_upper_we * s_track_relative_upper
            - f_e_upper_ew * s_track_relative_upper
            + P_region * (s_upper[t+1] / s_total)
        )[inner]

        # down and top: redistribute unaccounted water that is otherwise lost from the sytem
        lower_to_upper = np.maximum(0, s_track_lower - s_lower[t])
        upper_to_lower = np.maximum(0, s_track_upper - s_upper[t])
        s_track_lower[inner] = (s_track_lower - lower_to_upper + upper_to_lower)[inner]
        s_track_upper[inner] = (s_track_upper - upper_to_lower + lower_to_upper)[inner]

        # compute tracked evaporation
        e_track += evap[t] * (s_track_lower / s_lower[t+1])

        # losses to the north and south
        north_loss += (
            fy_n_upper_ns * s_track_relative_upper
            + fy_n_lower_ns * s_track_relative_lower
        )[1, :]

        south_loss += (
            fy_s_upper_sn * s_track_relative_upper
            + fy_s_lower_sn * s_track_relative_lower
        )[-2, :]

        east_loss += (
            f_e_upper_ew * s_track_relative_upper
            + f_e_lower_ew * s_track_relative_lower
        )[:, -2]

        west_loss += (
            f_w_upper_we * s_track_relative_upper
            + f_w_lower_we * s_track_relative_lower
        )[:, 1]

        # Aggregate daily accumulations for calculating the daily means
        s_track_lower_mean += s_track_lower / ntime
        s_track_upper_mean += s_track_upper / ntime

    # Pack processed data into new dataset
    ds = xr.Dataset(
        {
            # Keep last state for a restart
            "s_track_upper_restart": (["latitude", "longitude"], s_track_upper),
            "s_track_lower_restart": (["latitude", "longitude"], s_track_lower),
            "s_track_upper": (["latitude", "longitude"], s_track_upper_mean),
            "s_track_lower": (["latitude", "longitude"], s_track_lower_mean),
            "e_track": (["latitude", "longitude"], e_track),
            "north_loss": (["longitude"], north_loss),
            "south_loss": (["longitude"], south_loss),
            "east_loss": (["latitude"], east_loss),
            "west_loss": (["latitude"], west_loss),
        }
    )
    return (s_track_upper, s_track_lower, ds)


def initialize(config_file):
    """Read config, region, and initial states."""
    config = parse_config(config_file)
    region = load_region(config).values
    if config["restart"]:
        date = config["datelist"][-1] + pd.Timedelta(days=1)
        # Reload last state from existing output
        ds = xr.open_dataset(output_path(date, config))
        s_track_upper = ds.s_track_upper_restart.values
        s_track_lower = ds.s_track_lower_restart.values

    else:
        # Allocate empty arrays based on shape of input data
        s_track_upper = np.zeros_like(region)
        s_track_lower = np.zeros_like(region)

    return config, region, s_track_lower, s_track_upper


#############################################################################
# With the correct import statements, the code in the function below could
# alternatively be be used as a script in a separate python file or notebook.
#############################################################################


def run_experiment(config_file):
    """Run a backtracking experiment from start to finish."""
    config, region, s_track_lower, s_track_upper = initialize(config_file)

    for date in reversed(config["datelist"]):
        print(date)
        preprocessed_data = xr.open_dataset(input_path(date, config))

        # Resample to (higher) target frequency
        # After this, the fluxes will be "in between" the states
        fluxes, states = resample(preprocessed_data, config['target_frequency'])

        # Convert data to volumes
        change_units(fluxes, states, config["target_frequency"])

        # Apply a stability correction if needed
        stabilize_fluxes(fluxes, states)

        # Determine the vertical moisture flux
        fluxes["f_vert"] = calculate_fv(fluxes, states, config["periodic_boundary"], config["kvf"])

        (s_track_upper, s_track_lower, processed_data) = backtrack(
            date,
            fluxes,
            states,
            s_track_upper,
            s_track_lower,
            region,
            config,
        )

        # Write output to file
        # TODO: add (and cleanup) coordinates and units
        processed_data["longitude"] = preprocessed_data.longitude
        processed_data["latitude"] = preprocessed_data.latitude
        processed_data.to_netcdf(output_path(date, config))


###########################################################################
# The code below makes it possible to run wam2layers from the command line:
# >>> python backtrack.py path/to/cases/era5_2021.yaml
# or even:
# >>> wam2layers backtrack path/to/cases/era5_2021.yaml
###########################################################################


@click.command()
@click.argument('config_file', type=click.Path(exists=True))
def cli(config_file):
    """Run WAM2layers backtrack experiment from the command line.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - python path/to/backtrack.py path/to/cases/era5_2021.yaml
        - wam2layers backtrack path/to/cases/era5_2021.yaml
    """
    run_experiment(config_file)


if __name__ == "__main__":
    cli()



"""Generic functions useful for preprocessing various input datasets."""

import numpy as np
from scipy.interpolate import interp1d
import xarray as xr
import pandas as pd


# old, keep only for reference and ecearth starting case
def resample(variable, divt, count_time, method="interp"):
    """Resample the variable to a given number of timesteps."""

    _, nlat, nlon = variable.shape
    ntime = count_time * divt
    shape = (ntime, nlat, nlon)
    new_var = np.nan * np.zeros(shape)

    oddvector = np.zeros(ntime)
    partvector = np.zeros(ntime)
    for o in np.arange(0, ntime, divt):
        for i in range(1, divt):
            oddvector[o + i - 1] = (divt - i) / divt
            partvector[o + i] = i / divt

    for t in range(ntime):
        idx = int((t + 1) / divt + oddvector[t])

        if method == "bfill":
            new_var[t] = (1 / divt) * variable[idx - 1]
        elif method == "interp":
            new_var[t] = variable[idx - 1] + partvector[t] * (
                variable[idx] - variable[idx - 1]
            )
        else:
            raise ValueError(f"Unknown resample method {method}")

    return new_var


def get_grid_info(ds):
    """Return grid cell area and lenght of the sides."""
    dg = 111089.56  # [m] length of 1 degree latitude
    erad = 6.371e6  # [m] Earth radius

    latitude = ds.latitude.values
    longitude = ds.longitude.values
    grid_spacing = np.abs(longitude[1] - longitude[0])  # [degrees]

    # Calculate area TODO check this calculation!
    lat_n = np.minimum(90.0, latitude + 0.5 * grid_spacing)
    lat_s = np.maximum(-90.0, latitude - 0.5 * grid_spacing)

    a = np.pi / 180.0 * erad ** 2 * grid_spacing * abs(
        np.sin(lat_s * np.pi / 180.0) - np.sin(lat_n * np.pi / 180.0)
    )

    # Calculate faces
    ly = grid_spacing * dg  # [m] length eastern/western boundary of a cell
    lx_n_gridcell = ly * np.cos((latitude + grid_spacing / 2) * np.pi / 180)
    lx_s_gridcell = ly * np.cos((latitude - grid_spacing / 2) * np.pi / 180)
    lx = 0.5 * (lx_n_gridcell + lx_s_gridcell)
    return a, ly, lx


def join_levels(pressure_level_data, surface_level_data):
    """Combine 3d pressure level and 2d surface level data.

    A dummy value of 1100 hPa is inserted for the surface pressure
    """
    bottom = pressure_level_data.isel(lev=0).copy()
    bottom.values = surface_level_data.values
    bottom["lev"] = 110000.0  # dummy value

    # NB: plevs needs to come first to preserve dimension order
    return xr.concat([pressure_level_data, bottom], dim="lev").sortby(
        "lev", ascending=False
    )


def repeat_upper_level(pressure_level_data, fill_value=None):
    """Add one more level with the same value as the upper level.

    A dummy value of 100 hPa is inserted for the top level pressure
    """
    top_level_data = pressure_level_data.isel(lev=-1).copy()
    top_level_data["lev"] = 10000.0

    if fill_value is not None:
        top_level_data.values = np.ones_like(top_level_data) * fill_value

    return xr.concat([pressure_level_data, top_level_data], dim="lev")


def get_new_target_levels(surface_pressure, p_boundary, n_levels=40):
    """Build a numpy array with new pressure levels including the boundary."""
    # remove xarray labels if present
    surface_pressure = np.array(surface_pressure)

    pressure_upper = 20000
    dp = (surface_pressure - pressure_upper) / (n_levels - 1)

    levels = np.arange(0, n_levels)[None, :, None, None]
    ntime, _, nlat, nlon = surface_pressure.shape

    # Note the extra layer of zeros at the bottom and top of the array
    # TODO: find out why
    new_p = np.zeros((ntime, n_levels + 2, nlat, nlon))
    new_p[:, 1:-1, :, :] = surface_pressure - dp * levels

    mask = np.where(new_p > p_boundary, 1, 0)
    mask[:, 0, :, :] = 1  # bottom value is always 1

    # Insert the boundary in the new levels, pushing higher layers one index up
    # e.g. with the boundary at 850 hPa:
    # [0, 1000, 900, 800, 700, 600, 0] --> [0, 1000, 900, 850, 800, 700, 600]
    new_p[:, :-1, :, :] = (
        mask[:, 1:, :, :] * new_p[:, 1:, :, :]
        + (1 - mask[:, 1:, :, :]) * new_p[:, :-1, :, :]
    )

    new_p[:, 1:, :, :] = np.where(
        new_p[:, :-1, :, :] == new_p[:, 1:, :, :],
        p_boundary,
        new_p[:, 1:, :, :],
    )
    return new_p


def interpolate_old(old_var, old_pressure_levels, new_pressure_levels, type="linear"):
    """Interpolate old_var to new_pressure_levels."""
    new_var = np.zeros_like(new_pressure_levels)

    ntime, _, nlat, nlon = old_var.shape

    for t in range(ntime):
        for i in range(nlat):
            for j in range(nlon):

                pressure_1d = old_pressure_levels[t, :, i, j]
                var_1d = old_var[t, :, i, j]

                pressure_1d = pressure_1d[~pressure_1d.mask]
                var_1d = var_1d[~var_1d.mask]

                f_q = interp1d(pressure_1d, var_1d, type)
                new_var[t, :, i, j] = f_q(new_pressure_levels[t, :, i, j])

    return new_var


def interpolate(x, xp, fp, axis=1, descending=False):
    """Linearly interpolate along an axis of an N-dimensional array.

    This function interpolates one slice at a time, i.e. if xp and fp are 4d
    arrays, x should be a 3d array and the function will return a 3d array.

    It is assumed that the input array is monotonic along the axis.
    """
    # Cast input to numpy arrays
    x = np.asarray(x)
    xp = np.asarray(xp)
    fp = np.asarray(fp)

    # Move interpolation axis to first position for easier indexing
    xp = np.moveaxis(xp, axis, 0)
    fp = np.moveaxis(fp, axis, 0)

    # Handle descending axis
    if descending:
        xp = np.flip(xp, axis=0)
        fp = np.flip(fp, axis=0)
        assert np.diff(xp, axis=0).min() >= 0, "with descending=False, xp must be monotonically decreasing"
    else:
        assert np.diff(xp, axis=0).min() >= 0, "with desciending=True, xp must be monotonically increasing"

    # Check for out of bounds values
    if np.any(x[None, ...] < xp[0, ...]):
        raise ValueError("one or more x are below the lowest value of xp")
    if np.any(x[None, ...] > xp[-1, ...]):
        raise ValueError("one or more x are above the highest value of xp")

    # Find indices such that xp[lower] < x < xp[upper]
    upper = np.sum(x > xp, axis=0)
    lower = upper - 1

    # This will allow numpy advanced indexing to take an (N-1)D slice of an ND array
    upper = (upper, *np.meshgrid(*[range(l) for l in x.shape], indexing="ij"))
    lower = (lower, *np.meshgrid(*[range(l) for l in x.shape], indexing="ij"))

    fy = fp[lower] + (fp[upper] - fp[lower]) * (x - xp[lower]) / (xp[upper] - xp[lower])
    return fy


def insert_level(pressure_level_data, new_level, coord_value, dim_name="level"):
    """Insert a new level in the pressure level data.

    Note: new levels are inserted at the end of the dimension.
    Sorting by descending pressure is not performed automatically.

    Args:
        - pressure_level_data: xarray.DataArray with dimensions (time, lev, lat,
          lon)
        - new_level: the new data values that should be inserted as an
          additional level. If int or float, it will insert a constant value. If
          array like, it will insert the values in the array.
        - coord_value: coordinate value to use for the new level, e.g. 110000 Pa
          for well below the surface. Must be unique.
    """
    # Create dummy array that can be concatenated with the original data.
    dummy = pressure_level_data.isel({dim_name: 0}).copy()
    dummy[dim_name] = coord_value

    # Insert the nedim_name data into the dummy array
    if isinstance(new_level, xr.DataArray):
        dummy.values = new_level.values
    elif isinstance(new_level, np.ndarray):
        dummy.values = new_level
    elif isinstance(new_level, (int, float)):
        dummy.values = np.ones_like(dummy) * new_level
    else:
        raise ValueError("Invalid type for new_level")

    return xr.concat([pressure_level_data, dummy], dim=dim_name)


def sortby_ndarray(array, other, axis):
    """Sort array along axis by the values in another array."""
    idx = np.argsort(other, axis=axis)
    return np.take_along_axis(array, idx, axis=axis)


def calculate_humidity(dewpoint, pressure):
    """
    Calculate the specific humidity from (surface) pressure and
    dew point temperature

    See further details at eq. 7.4 and 7.5 (Page 102) of:
    https://www.ecmwf.int/en/elibrary/20198-ifs-documentation-cy47r3-part-iv-physical-processes
    """
    Rd = 287.0597
    Rv = 461.5250
    a1 = 611.21
    a3 = 17.502
    a4 = 32.19
    t0 = 273.15

    # Calculation of saturation water vapour pressure from Teten's formula
    svp = a1 * np.exp(a3 * (dewpoint - t0) / (dewpoint - a4))

    # Specific humidity
    spec_hum = (Rd / Rv) * svp / (pressure - ((1 - Rd / Rv) * svp))

    return spec_hum


def accumulation_to_flux(data):
    """Convert precip and evap from accumulations to fluxes.

    Incoming data should have units of "m"

    Note: this does not currently modify the time labels. It can be argued
    that the times should be shifted. A good fix for this is welcome.
    """
    density = 1000  # [kg/m3]
    timestep = pd.Timedelta(data.time.diff('time')[0].values)
    nseconds = timestep.total_seconds()

    fluxdata = (density * data / nseconds).assign_attrs(units="kg m-2 s-1")

    # TODO: Adjust time points?
    # original_time = data.time
    # midpoint_time = original_time - timestep / 2
    # data["time"] = midpoint_time

    # Extrapolation introduces a small inconsistency at the last midnight...
    # data.interp(time=original_time, kwargs={"fill_value": "extrapolate"})
    return fluxdata



from pathlib import Path

import click
import matplotlib.pyplot as plt
import xarray as xr
from cartopy import crs
from cartopy import feature as cfeature
from wam2layers.preprocessing.shared import get_grid_info
from wam2layers.tracking.backtrack import (input_path, load_region,
                                           output_path, parse_config)


def polish(ax, region):
    ax.add_feature(cfeature.COASTLINE, linewidth=0.8)
    ax.add_feature(cfeature.BORDERS, linestyle="-", linewidth=0.2)
    ax.set_xlim(region.longitude.min(), region.longitude.max())
    ax.set_ylim(region.latitude.min(), region.latitude.max())


def visualize_input_data(config_file):
    raise NotImplementedError()
    config = parse_config(config_file)
    for date in config["datelist"]:
        ds = xr.open_dataset(input_path(date, config))
    # do stuff


def visualize_output_data(config_file):
    raise NotImplementedError()
    config = parse_config(config_file)
    for date in config["datelist"]:
        ds = xr.open_dataset(output_path(date, config))
    # do stuff


def visualize_both(config_file):
    """Diagnostic figure with four subplots combining input and output data."""
    config = parse_config(config_file)
    region = load_region(config)
    a_gridcell, lx, ly = get_grid_info(region)

    out_dir = Path(config["output_folder"]) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)

    for date in config["datelist"]:
        print(date)
        ds_in = xr.open_dataset(input_path(date, config))
        ds_out = xr.open_dataset(output_path(date, config))

        # Make figure
        fig, [[ax1, ax2], [ax3, ax4]] = plt.subplots(2, 2, subplot_kw=dict(projection=crs.PlateCarree()), figsize=(14,8))

        ax1.set_title("Tracked precipitation" + date.strftime("%Y%m%d"))
        precip = ds_in.precip.sum('time') * region / a_gridcell[:, None] * 1000
        precip.plot(ax=ax1, cmap='Blues')
        polish(ax1, region)

        ax2.set_title("Moisture source")
        e_track = ds_out.e_track / a_gridcell[:, None] * 1000
        e_track.plot(ax=ax2, cmap='GnBu')
        polish(ax2, region)

        ax3.set_title("S track upper layer")
        s_track_upper = ds_out.s_track_upper / a_gridcell[:, None] * 1000
        s_track_upper.plot(ax=ax3, cmap="YlOrRd")
        ds_in.mean('time').plot.streamplot(x="longitude", y="latitude", u="fx_upper", v="fy_upper", ax=ax3, color="black")
        polish(ax3, region)

        ax4.set_title("S track lower layer")
        s_track_lower = ds_out.s_track_lower / a_gridcell[:, None] * 1000
        s_track_lower.plot(ax=ax4, cmap="YlOrRd")
        ds_in.mean('time').plot.streamplot(x="longitude", y="latitude", u="fx_lower", v="fy_lower", ax=ax4, color="black")
        polish(ax4, region)

        # Save
        output_file = out_dir / f"input_output_{date.strftime('%Y%m%d')}.png"
        fig.savefig(output_file)
        plt.close()

###########################################################################
# The code below makes it possible to run wam2layers from the command line:
# >>> python visualization.py input path/to/cases/era5_2021.yaml
# or even:
# >>> wam2layers visualize output path/to/cases/era5_2021.yaml
###########################################################################


@click.group()
def cli():
    """Visualize input or output data of a WAM2layers experiment"""
    pass


@cli.command()
@click.argument('config_file', type=click.Path(exists=True))
def input(config_file):
    """Visualize input data for experiment."""
    visualize_input_data(config_file)


@cli.command()
@click.argument('config_file', type=click.Path(exists=True))
def output(config_file):
    """Visualize output data for experiment."""
    visualize_output_data(config_file)


@cli.command()
@click.argument('config_file', type=click.Path(exists=True))
def both(config_file):
    """Visualize both input and output data for experiment."""
    visualize_both(config_file)

if __name__ == "__main__":
    cli()



import click
from wam2layers.preprocessing.era5 import cli as era5_cli


@click.group()
def cli():
    """Pre-process raw input data for tracking with WAM2layers """
    pass


cli.add_command(era5_cli, name="era5")
